(* ******************************************************************** *)
(*  									*)
(* 	Project		: Ida - Isabelle's documentation assistant	*)
(*  	Author 		: B. Wolff					*)
(*  	Affiliation 	: University of Bremen, 			*)
(*			  Bremen Institute of Safe Systems (BISS)	*)
(*   	This file	: Documentation puncher				*)
(*			  General mechanism for protocolling proofs in  *)
(*			  a docfile					*)
(*  									*)
(*  	$Date: 1997/09/13 17:19:32 $					*)
(*      $Revision: 1.22 $						*)
(*  									*)
(*  ******************************************************************* *)



infix 5 $$;


signature PUNCH_LANG =
  sig
    val indent_depth            : int
    val right_margin_init       : int
    val left_margin_init        : int

    val short_num_bracket_left  : string
    val short_num_bracket_right : string
    val symb_subprf_refine      : string
    val symb_subprf_equiv       : string

    val blanks			: int * int -> int -> string
    val newline			: int -> int * int -> string
  end;


signature PUNCHER =
  sig
    exception PUNCH of string


 (* external signature data-base management *)
    val syn_db                  : Syntax.syntax Symtab.table ref
    val syn_db_update           : string * Syntax.syntax -> unit
    val get_external_syntax     : theory -> Syntax.syntax
    (* coerces internal syntax according syn_db and user data: *)
    val create_ext_syntax       : theory-> (string*typ*mixfix)list
					-> (string*string*mixfix)list
					-> Syntax.syntax
    (* creates and updates: *)
    val attach_ext_syntax       : theory-> (string*typ*mixfix)list
					-> (string*string*mixfix)list
					-> unit    


 (* init and control punching *)
    val set_right_margin        : int -> unit
    val set_left_margin         : int -> unit
    val open_docfile            : string -> unit
    val close_docfile           : unit -> unit
    val doc_on                  : bool ref (* punching on/off *)

    (* alternative use for items and consecutive proof doc generation:*)
    val punch_doc_item          : string -> Ref_mgr.docit -> string -> unit
    val punch_proof_item_end    : string -> string -> unit
    val punch_proof_item_start  : unit -> unit



 (* structured comments *)
    datatype cmt =    Ref of string (* References to theorem names *)
                    | Cmt of string (* Just text (must not contain \n!!!) *)
                    | frm of term   (* Formula as term in internal syntax *)
                    | tym of typ    (* Type as typ in internal syntax *)
                    | Tab	    (* Tab *)
                    | Brk	    (* Break *)
                    | Block of 	    (* Comment block *)
			cmt list
    val Frm	  		: string -> cmt (*string version (intern. syn)
						  using topthm() *)
    val Tym	  		: string -> cmt (*string version (intern. syn)
						  using topthm() *)
    val cmt_to_text             : theory -> cmt list -> string

 

 (* basic docticals; only applicable if "punch_proof_item_start" done *)
    val $$        : 'a * (thm -> tactic) -> tactic  (* forcing tactics *)
    val CMT       : cmt list -> int -> tactic
    val NR_CMT    : cmt list -> int -> tactic
    val CMTGOAL   : cmt list * cmt list -> int -> tactic
    val PUNCHGOAL : bool list  -> cmt list -> int -> tactic


 
 (* punching other objects *)
    val punch_thm_obj : string * string * Ref_mgr.docit_kind * 
			Ref_mgr.docit_cite_style * (int*int*int) * thm -> unit			 
    val punch_thm : string * thm -> unit
    val punch_def : string * thm -> unit
    val punch_thy : string * theory -> unit
    val punch_sig : string * theory -> unit


    val string_of_term : theory-> term-> string

 (* re-export *)
    val symb_subprf_refine : string
    val symb_subprf_equiv  : string

  end



functor Doc_punch  (M : PUNCH_LANG) : PUNCHER =
struct

local open M Syntax Num_mgr Ref_mgr in

exception PUNCH of string;

val symb_subprf_refine = M.symb_subprf_refine;
val symb_subprf_equiv  = M.symb_subprf_equiv;

    

(* *********************************************************************** *)
(*									   *)
(* External syntax data-base management					   *)
(*									   *)
(* *********************************************************************** *)

fun get_thyname thy =
    ! (hd (stamps_of_thy thy))
    (* this gets rid of unnamed theories. 
    hd (filter (fn nm=> not ("#"= nm)) (map ! (stamps_of_thy thy)))
    handle LIST _ => ""  *)

(* ord refl & trans. a ord b == a <= b *)
fun topsort ord [] = []
  | topsort ord (a::[]) = [a]
  | topsort ord (a::S)  = if exists ((curry ord) a) S then topsort ord S
			  else a :: (topsort ord S);



val syn_db =  ref(Symtab.null:(Syntax.syntax Symtab.table));

fun syn_db_update (str,syn) = (syn_db := (Symtab.update((str,syn),(!syn_db))));

fun get_infix_repr n syn = Syntax.infixes_of n syn;

fun create_ext_syntax0 fl thy ty_li li =
(* takes syntax syn of thy, combines all parent syntaxes with syntaxes from 
   the syn_db if available and overwrites all definitions in syn by this 
   combination. Finally updates the overwrite with additional increments li.

   This way a syntax built by use_thy can be coerced
   under full exploitation of knowledge out the data-base and
   explicit signature declarations given by the user.
*)
    let val thy_nam = get_thyname thy;
        val max_par = parents_of thy;
	fun lum y   = case Symtab.lookup(!syn_db,get_thyname y) of
			Some x => x
		      | _      => (warning ("Syntax for "^ (get_thyname y) ^ 
				  " not in extern signature data-base"); 
				  create_ext_syntax0 fl y [] [])
	val par_syns= map lum max_par;
       	fun par_syn()= foldl (uncurry Syntax.merge_syntaxes) 
				(hd par_syns, tl par_syns);
        val syn     = syn_of thy
 	fun conv (s,ty,mi) = (s,typ_of(read_ctyp (sign_of thy) ty),mi)
        fun tra_par () = (writeln "transfering external into internal syntax";
			  if null max_par then  syn
			  else	transfer_prtab_in_syntax (par_syn()) syn)
	val new_syn = Syntax.update_prtab_in_syntax (tra_par ())
				(ty_li @ (map conv li))
    in  if fl then syn_db_update(get_thyname thy,new_syn) else ();
    	new_syn
    end;

fun create_ext_syntax thy ty_li li = create_ext_syntax0 false thy ty_li li

fun attach_ext_syntax thy ty_li li = ignore (create_ext_syntax0 true thy ty_li li)


fun get_external_syntax thy =
    let val thy_nam = get_thyname thy;
	val newSyntax = Symtab.lookup(!syn_db,thy_nam)
    in  case newSyntax of 
	None => raise PUNCH("External syntax for: " ^ thy_nam ^ " not defined!")
       |Some X => X
    end;

(* *********************************************************************** *)
(*									   *)
(* Long Name Expansion							   *)
(*									   *)
(* *********************************************************************** *)

(* At the end of a proof doc generation, a dictionary for all Name-references
   is produced that associates with each external name of a Name-reference
   a "long name", i.e. "Definition Natural Numbers" etc... according to the
   style-setting in the Ref_mgr.
 *)

fun gen_long_names() = (); (* NOT YET IMPLEMENTED !!! *)


(* *********************************************************************** *)
(*									   *)
(* Basic punching layer in documentation file				   *)
(*									   *)
(* *********************************************************************** *)

val docfile   = ref(std_out);
val doc_on    = ref(false);
val proof_on  = ref(false);
val doc_it_no = ref(0);


fun open_docfile l      = (docfile := open_out (l^".m4"); doc_on:=true;
			   output (!docfile,"changequote([****, ****])");
			   doc_it_no:=0;
			   clear_ref_db ());
fun write_doc str       = output (!docfile, str);
fun close_docfile ()    = (gen_long_names();
			   (* write_doc ">>CUT_HERE<<\n"; *)
			   (* Marker for external merge-script*)
			   doc_on:=false;proof_on:=false;
			   close_out(!docfile));

(* The proof generation channel can either be opened for proof-documentation
   (sequential, doctical controlled mode) or direct punching of documentation
   items (theories, signatures, etc). The kind of openess is controlled via
   proof_on. Violations lead to exceptions... *)


fun punch_doc_item name docit str = 
    if !doc_on then
       if !proof_on then raise PUNCH "punch_object during proof-punching..."
       else (update_ref_db(name,docit);
	     doc_it_no := (!doc_it_no) + 1;
	     write_doc ("define("^(get_extern_name name)^", [**** ");
	     write_doc str;
	     write_doc " ****])")
    else ();
	 

fun punch_proof_item_start () = 
    if !doc_on then
	if !proof_on then raise PUNCH "proof-punching during proof-punching..."
        else 	(proof_on:=true;init_num();
	         write_doc ("define(INTIT_"^(makestring (!doc_it_no))^
				", [****[****")
		 (* punching into internal object since final name not yet
                    available ... *)
	        )
    else ();

fun docit sty ki lo_na thy_name = 
	Docit{  long_name   = lo_na,
		thy_name    = thy_name,
		short_name  = get_docstate(),
		kind        = ki,
		cite_style  = sty};


fun punch_proof_item_end name thy_name =
    if !doc_on then
	if not (!proof_on) then raise PUNCH "proof-punching already closed!"
        else 	 (update_ref_db(name,docit short prf name thy_name);
		 doc_it_no := (!doc_it_no) + 1;
                 (* ... creating ref *)
		 write_doc " ****]****])";
                 (* ... closing internal object *)
	 	 write_doc ("define(" ^ (get_extern_name name) ^"," ^
				"INTIT_" ^ (makestring (!doc_it_no - 1)));
                 (* ... defining ref by internal object *)
	         write_doc ")"; (* ... close *)
                 proof_on:=false; init_num())
    else ();



(* *********************************************************************** *)
(*									   *)
(* Formatting layer (just text generation)				   *)
(*									   *)
(* *********************************************************************** *)

val right_margin = ref(right_margin_init);
val left_margin  = ref(left_margin_init);

fun set_right_margin n = (right_margin:=n);
fun set_left_margin  n = (left_margin:=n);


fun short_number () = 
	short_num_bracket_left ^
	(makestring(indent_level ())) ^
	short_num_bracket_right ^
	(makestring (top_nums()))
	^ ". ";


fun doc_string_of_typ n syn t = 
	(Pretty.setblanks M.blanks;
	Pretty.setnewline (M.newline n);
	Pretty.setmargin ((!right_margin) - (!left_margin) - n);
	(* settings for internal breaks *)
	(let val t = Syntax.string_of_typ syn t
	in  Pretty.resetblanks(); 
	    Pretty.resetnewline();
	    Pretty.setmargin 80;    (* dangerous !!! *)
	    t 
        end));


fun doc_string_of_term n syn t = 
	(Pretty.setblanks M.blanks;
	Pretty.setnewline (M.newline n);
	Pretty.setmargin ((!right_margin) - (!left_margin) - n);
	(* settings for internal breaks *)
	(let val t = Syntax.string_of_term true syn t
	in  Pretty.resetblanks(); 
	    Pretty.resetnewline();
	    Pretty.setmargin 80; (* dangerous !!! *)
	    t 
        end));

fun string_of_term thy t =
    let val thy_nam = get_thyname thy
	val newSyntax = Symtab.lookup(!syn_db, thy_nam)
    in
	case newSyntax of 
	    None => raise PUNCH("External syntax for: " ^ thy_nam 
				^ " not defined!")
	  |  Some newSyntax => doc_string_of_term 0 newSyntax t
    end


(* Fehlt : Typ,term,and blocks... !!! *)
datatype cmt   = Ref of string 
               | Cmt of string

               | frm of term
               | tym of typ

               | Block of cmt list
               | Tab
               | Brk;

fun ind_level () = ((Num_mgr.indent_level()) -1) * indent_depth;

(* The breaks were performed according to the indentlevel !!! *)
fun  cmt_to_text thy [] = ""
   | cmt_to_text thy ((Ref s):: R)= 
		let val na = (get_extern_name s) 
			     handle OPTION (_) => s
		in  na ^" "^(cmt_to_text thy R) end
   | cmt_to_text thy ((Cmt s):: R)= s ^ (cmt_to_text thy R)
   | cmt_to_text thy ((frm tm):: R)= 
		let val s' = doc_string_of_term (ind_level()) 
				(get_external_syntax thy) tm
                in  " "^s' ^" "^ (cmt_to_text thy R) end
   | cmt_to_text thy ((tym t) :: R)= 
		doc_string_of_typ (ind_level())(get_external_syntax thy) t
   | cmt_to_text thy (Brk :: R)= 	
		(M.newline(ind_level())(0,0)) ^ cmt_to_text thy R
   | cmt_to_text thy (Tab :: R)= 	
		(M.blanks(0,0)(indent_depth)) ^ cmt_to_text thy R
   | cmt_to_text thy ((Block R) :: R')= (push_nums ();
					 let val S = cmt_to_text thy R
                                             val S'=(pop_nums ();
						     cmt_to_text thy R')
                                         in  S ^ S' end);	


fun  Frm s = frm(term_of(read_cterm(sign_of 
		(theory_of_thm (topthm())))(s,HOLogic.boolT)));

fun  Tym s = tym(typ_of(read_ctyp(sign_of 
		(theory_of_thm (topthm())))(s)));



(* the string should be scanned for \n and roughly formatted. 
   NOT YET IMPLEMENTED !!! *)


   

fun doc_sig thy = 
    let val par = parents_of thy;
        fun strip_consts thy = map fst (Symtab.dest(#const_tab 
					(Sign.rep_sg (sign_of thy))));
        val par_consts = flat(map strip_consts par);
        val thy_consts = (strip_consts thy);
        val loc_consts = foldl (op \) (thy_consts, par_consts);
        val infix_tab  = Symtab.make (get_infix_repr 1 
						(get_external_syntax thy));
	val const_tab  = #const_tab (Sign.rep_sg (sign_of thy));
        fun typ_2 m    = Type.unvarifyT(the (Symtab.lookup(const_tab,m)));
        fun expand m   = case Symtab.lookup(infix_tab,m) of
			   None   => (m, typ_2 m)
			 | Some x => (x, typ_2 m);
        val loc_sig    = sort (fn ((a,_),(b,_))=> a<b) (map expand loc_consts);
        val max_par    = topsort subthy par;
    in  (map get_thyname max_par, loc_sig) end;



(* *********************************************************************** *)
(*									   *)
(* Fundamental doctical combinators					   *)
(*									   *)
(* *********************************************************************** *)


fun (_ $$ f) = prune_params_tac THEN (STATE f); (* COUNTER HACK !!! *)


fun NR_CMT' str n top = 
    if !doc_on then 
	(write_doc(M.blanks(0,0)(ind_level()));
	write_doc(short_number());
	write_doc(cmt_to_text (theory_of_thm top) str);
	write_doc(M.newline(0)(0,0)); 
	all_tac)
    else all_tac;

fun CMT' str n top = 
    if !doc_on then 
       (write_doc(M.blanks(0,0)(ind_level()));
	write_doc(cmt_to_text (theory_of_thm top) str);
	write_doc(M.newline(0)(0,0)); 
	all_tac)
    else all_tac;

fun NR_CMT str n = ()$$(NR_CMT' str n);
fun CMT str n = ()$$(CMT' str n);



fun PUNCHGOAL' with_prems str n top = 
(* with_prems = [false] should prune printing of potential first premisses - 
   NOT YET IMPELEMENTED !!! *)
    if !doc_on then (
    let val thy_nam = get_thyname (theory_of_thm top)
	val newSyntax = Symtab.lookup(!syn_db,thy_nam);
        fun getgoal i = (case  drop (i-1, prems_of top)  of
	    [] => raise PUNCH"getgoal: Goal number out of range"
	  | Q::_ => Q);


	val t =  case newSyntax of 
			None => raise PUNCH("External syntax for: " ^ thy_nam 
					^ " not defined!")
		     |  Some newSyntax =>doc_string_of_term
						(ind_level()) 
						newSyntax 
						(getgoal n)
    in  (by(prune_params_tac); (* COUNTER-HACK *)
	 write_doc(M.blanks(0,0)(ind_level()));
	 write_doc(cmt_to_text (theory_of_thm top) str);(* Leading cmt ... *)
	 write_doc t;
	 write_doc(M.newline(0)(0,0));
	 all_tac)
    end)
    else all_tac;

fun PUNCHGOAL with_prems str n = ()$$(PUNCHGOAL' with_prems str n);

fun CMTGOAL' (str,str') n top = EVERY[NR_CMT str n, PUNCHGOAL [] str' n];
fun CMTGOAL  str n            = ()$$(CMTGOAL' str n);


(* *********************************************************************** *)
(*									   *)
(* Punch operations							   *)
(* (Formatting is always depending on the ind_level !!!)		   *)
(*									   *)
(* *********************************************************************** *)

val dummy = (0,0,0);
fun freeze thm = fst(freeze_thaw thm);

fun punch_thm_obj (n , name, ki, sty, sh_na, t) = 
    if !doc_on then
       let val extSyntax   = get_external_syntax (theory_of_thm t);
           val docit       = Docit {cite_style=sty,kind=ki,
	  			short_name=sh_na,long_name=name,
				thy_name=get_thyname (theory_of_thm t)}
	   val tt          = (doc_string_of_term (ind_level()) extSyntax 
			  (#prop (rep_thm (freeze t))))
       in  punch_doc_item n docit tt end
    else ();



fun punch_thm (n , t)   = punch_thm_obj (n,n,theorem,long,dummy,t);

fun punch_def (n , t)   = punch_thm_obj (n,n,def,long,dummy,t);

fun mk_head thy_nam pars=
    let val pars'	= foldr (op ^) ((map (fn a => a^" +") pars),"")
    in  [Tab,Cmt thy_nam, Cmt" = ", Cmt pars', Brk] end;

fun mk_sig_body sign    = 
    (* speculative indents for the type part for the moment ... *)
    Tab::flat(map (fn (na,ty) =>[Cmt na,Tab,Tab,
			    Block[Block[Block[Cmt ":",tym ty]]],Brk]) sign);
  
fun mk_thm (na,t)     = [Ref na,Cmt ":", Tab,Tab,
			   Block[Block[Block[Block[
				frm(#prop(rep_thm(freeze t)))]]]],Brk];

fun punch_sig (name,thy)= if not(!doc_on) then () else
   (let val thy_nam     = get_thyname thy;
        val (pars,sign) = doc_sig thy;
	val head        = mk_head thy_nam pars;   
	val body	= mk_sig_body sign;
	val tt		= cmt_to_text thy [Block (head @ [Block body])];
	val docit       = Docit {cite_style=long,
				 kind=sig_class,short_name=dummy,
				 long_name=name,thy_name=thy_nam}
    in  punch_doc_item name docit tt end);



fun punch_thy (name,thy)= if not(!doc_on) then () else
   (let val thy_nam     = get_thyname thy;
        val (pars,sign) = doc_sig thy;
	val head        = Brk::(mk_head thy_nam pars);   
	val sig_body	= [Block(Tab::mk_sig_body sign)];
	val axm_head    = [Brk, Cmt "Definitions:", Brk]
	val axm_body    = [Block(flat (map mk_thm (axioms_of thy)))]
	val thm_head    = [Brk, Cmt "Theorems:", Brk]
	val thm_body    = [Tab,Block(flat (map mk_thm (thms_of thy)))]
	val cmtt        = [Block(head @ sig_body @ 
				axm_head @ axm_body @ 
				thm_head @ thm_body)]
	val tt 		= cmt_to_text thy cmtt;
	val docit       = Docit {cite_style=long,
				 kind=thy_class,short_name=dummy,
				 long_name=name,thy_name=get_thyname thy}
    in  punch_doc_item name docit tt end);


end;

end;
