(* ******************************************************************** *)
(*  									*)
(* 	Project		: Ida - Isabelle's documentation assistant	*)
(*  	Author 		: B.Wolff					*)
(*  	Affiliation 	: University of Bremen, 			*)
(*			  Bremen Institute of Safe Systems (BISS)	*)
(*   	This file	: Dictionary for external and internal 		*)
(*			  documentation references + info		*)
(*  									*)
(*  	$Date: 1997/01/22 19:48:36 $					*)
(*      $Revision: 1.5 $						*)
(*  									*)
(*  ******************************************************************* *)


(* The general syntax for external names (i.e.thase names that can be used
as references in external document templates) is:

	<kind><thyname>_<name>

where kind see below. "_<name>" is ommited for signatures and theories.
The kind "name" is used to refer to references themselves.
(useful in connection with docit_cite_style = long).

There may be several objects in the data-base to the same base-name.
Eg. some prf, sum and the theorem under the name "bla". Their
external name will distinguish them, but internally they are identical.
Hence, the name-docit relation is i.g. not a function and
symtab can only be used with a trick: when storing them in the data-base,
appropriate blanks " " will be suffixed to a name. They dont hurt 
during the external name generation, internally it is always clear
when to add blanks, so there was no use to muddle with list's of
docit's around...

*)

structure Ref_mgr =

struct

   datatype docit_kind = theorem | lemma | corollary | induct 
		        | axiom | def | prf | prfsum 
		        | type_class | term_class | sig_class | thy_class 
			| name_ref;

   datatype docit_cite_style = long | short;

   datatype docit = Docit of 
		 {	cite_style  : docit_cite_style,
			kind	    : docit_kind,
			short_name  : int * int * int,
			long_name   : string,
			thy_name    : string
		 };
		

   val empty_ref_db = Symtab.null : docit Symtab.table;

   val ref_db = ref (empty_ref_db);
   fun update_ref_db (s,dc) = (ref_db := Symtab.update((s,dc), !ref_db));
   fun lookup_ref_db (s,dc) = (ref_db := Symtab.update((s,dc), !ref_db));
   fun clear_ref_db () = (ref_db := empty_ref_db); 

   fun conv ch = if is_letter ch orelse is_digit ch orelse ch = "_" 
		then ch
		else "HEX" ^makestring(ord ch);


   fun get_extern_name refce = 
      let  val Docit({kind,thy_name, ...}) = the (Symtab.lookup(!ref_db,refce));
           fun conv ch = if is_letter ch orelse is_digit ch orelse ch = "_" 
		then ch
		else "HEX" ^makestring(ord ch);
           fun strip_blanks [] = []
              |strip_blanks (" "::S) = strip_blanks S
              |strip_blanks S = S
           fun strip_blank_suffix S = rev (strip_blanks (rev S))
           val refce = implode (map conv (strip_blank_suffix (explode refce)))
      in   case kind of 
 		  theorem	=> "DOCIT_thm_"^thy_name^"_"^refce
		| lemma 	=> "DOCIT_thm_"^thy_name^"_"^refce
		| corollary 	=> "DOCIT_thm_"^thy_name^"_"^refce
		| induct 	=> "DOCIT_thm_"^thy_name^"_"^refce
		| axiom  	=> "DOCIT_axm_"^thy_name^"_"^refce
		| def 	 	=> "DOCIT_def_"^thy_name^"_"^refce
		| prf 	 	=> "DOCIT_prf_"^thy_name^"_"^refce
		| prfsum 	=> "DOCIT_sum_"^thy_name^"_"^refce
		| type_class 	=> "DOCIT_typ_"^thy_name^"_"^refce
		| term_class 	=> "DOCIT_trm_"^thy_name^"_"^refce
		| sig_class 	=> "DOCIT_sig_"^thy_name
		| thy_class 	=> "DOCIT_thy_"^thy_name
		| name		=> "DOCIT_nam_"^thy_name^"_"^refce
      end;

    fun gen_name refce = 
	let  val Docit({cite_style,kind,short_name,long_name,...}) = 
					the(Symtab.lookup(!ref_db,refce));
	   val (c,p,sp) = short_name
	   val sh_na    = (makestring c) ^ 
				(if p > 0 then "."^(makestring p) ^
				        (if sp > 0 then "." ^(makestring sp)
					 else "")
				 else "");
	   val head     = (case kind of
		 	  theorem	=> "Theorem "
			| lemma 	=> "Lemma "
			| corollary 	=> "Corollary "
			| induct 	=> "Induction Scheme "
			| axiom  	=> "Axiom "
			| def 	 	=> "Definition "
			| prf 	 	=> "Proof "
			| prfsum 	=> "Subproof "
			| _ 		=> "")
	in   case cite_style of 
		  long  => head ^ sh_na ^ ":" ^ long_name
		| short => sh_na
        end;

   fun mapp f [] = ()
      |mapp f (a::S) = (f a; mapp f S);


   fun get_refs_of_ref_db () = 
	sort (op <= ) (map (get_extern_name o fst) 
		(Symtab.dest (!ref_db)));

end;
