(* ***************************************************************************
 
   $Source: /home/uniform/rep/isa_gui/tas/api.sml,v $
 
   A (fairly basic) API for TAS.
 
   This just wraps up the TAS external datatypes in a nice way. The API is 
   used from within the TAS image. You first open TAS -- mibbe this could
   be done before the TAS image is dumped? -- then use TasAPI.init() to start
   TAS with its initial state. On finishing, TAS will return a string which 
   if given as an argument to the function TasAPI.go restarts TAS where 
   it left off.
  
   $Date: 1997/09/11 15:40:14 $
   $Revision: 1.10 $
   Author: cxl (Last modification by $Author: cxl $)

   (C) 1997, Bremen Institute for Safe Systems, Universitaet Bremen
 
  ************************************************************************** *)




signature TAS_API =
    sig 
	include TAS_OBJECTS 

	(* start TAS in its initial state *)
	val start: unit-> unit

	(* resume TAS at a given state *)
	val resume : state-> unit
	    
	(* like above, but dump the state into file rather than std_out *)
	val startfile  : string-> unit
	val resumefile : string-> state-> unit 

	(* Set the directory where external objects can be found. 
	 * Be careful that this is set to the same directory when restarting
	 * TAS; hence this should actually be part of the configuration. *)
	val set_external_dir : string-> unit

    end


functor TasAPI(structure TAS: TAS) : TAS_API = 

struct


    open TAS 

    local

    (* status/error messages *)
    fun tas_msg msg = IO.output(IO.std_err, "+++ TAS: "^msg^"\n")

    exception TAS_Abort

    infix ^^
	
    fun x ^^ y     = x^","^y
    fun prt2 x y   = "("^x^^y^")"
    fun prt3 x y z = "("^x^^y^^z^")"

    val toString   = concat o (map Char.toString) o Lib.String.explode

    fun q str = "\""^(toString str)^"\""

    fun prt_list ls   = "["^(BasicUtil.concatWith "," ls)^"]"

    fun prt_coord (x, y) = prt2 (Int.toString x) (Int.toString y)

    fun prt_subst sub= prt_list (map (fn(x, y)=> prt2 (q x) (q y)) sub)

    fun prt_trafo t = rep t

    fun prt_op (opApplyTrafo(trf, inst)) = 
	         "opApplyTrafo"^(prt2 (prt_trafo trf) (prt_list (map q inst)))
      | prt_op (opApplyTrafoAt(trf, inst, p))=
	         "opApplyTrafoAt"^(prt3 (prt_trafo trf) (prt_list (map q inst)) (q p))
      | prt_op (dischargeThm(thm, int))=
	         "dischargeThm"^(prt2 (q "<some theorem>") (Int.toString int))

(*  old line, causes compiler to crash:
    fun dump_obj (Dev(th, str, (ix, ops), inf)) =
	         "Dev("^"thy"^^str^^
	            (prt2 ix (prt_list (map prt_op ops)))^^inf^")"  *)

    fun prt_hist (ix, ops)= prt2 (Int.toString ix) (prt_list (map prt_op ops))

    fun dump_obj (Dev(th, sp, h, inf)) =
	"Dev("^(q th)^^(q sp)^^(prt_hist h)^^(q inf)^")" 
      | dump_obj (TrafoObj(trf, sub)) =
	"TrafoObj"^(prt2 (prt_trafo trf) (prt_subst sub))
      | dump_obj (Subst(sub))=
	"Subst("^(prt_subst sub)^")"
      | dump_obj (Sign(th))=
	"Sign("^(q th)^")"
      | dump_obj (Spec(sp))=
	"Spec("^(q sp)^")"
   
    fun dump_obj_rep (obj, nm, wh)= prt3 (dump_obj obj) (q nm) (prt_coord wh)
	
    fun dump_state out state =
	(Debug.warn_off(); 
	 output(out, ("["^(BasicUtil.concatWith ",\n " 
			   (map dump_obj_rep state))^"]\n"));
	 tas_msg("state written")
	)

    fun tas_abort () = tas_msg "user abort"
	             
    fun tas_exit(out, path) () =
	(close_out out;
	 tas_msg ("exited, state saved in "^path))

    fun open_outfile path =
	(OS.FileSys.remove path handle (OS.SysErr _) => () ;
	     (* remove file if exists but we dinnae care if it disnae *)
	 IO.open_out path)

    in
		 
    fun start () = 
	TAS.go(initial_state(), dump_state std_out, K(), tas_abort)

    fun startfile path = 
	let val out= open_outfile path
	in  TAS.go(initial_state(), dump_state out, tas_exit(out, path), 
		    tas_abort)
	end handle OS.SysErr(str, _) => tas_msg ("I/O error, "^str)

    fun resume state = 
	TAS.go(state, dump_state std_out, K(), tas_abort)   

    fun resumefile path state=
	let val out= IO.open_out path
	in  TAS.go(state, dump_state out, tas_exit(out, path), tas_abort)
	end handle OS.SysErr(str, _) => tas_msg ("I/O error, "^str)

    fun set_external_dir dir = 
	(TAS.external_dir := dir; loadpath := dir :: (!loadpath))
	
    end	   
end  


