(* ***************************************************************************

   $Source: /home/uniform/rep/isa_gui/tas/trafos.sml,v $
 
   Generic Transformations Made Easy. 

   TRAFOTHEORY is the signature for an structure providing a
   "transformation theory". This consists of a theory in which the
   transformations live, some theorems saying that the transformation
   relation is transitive, reflexive and a congruence; and a list of
   all transformations which are (initially) known.
   
   TRAFOS is the export interface of the functor Trafos, which takes 
   such a theory and produces the necessary tactics and ML functions
   to handle transformations, application of transformations, discharging
   of proof obligations and all that. 

   $Date: 1997/09/14 18:00:55 $
   $Revision: 1.22 $
   Author: cxl (Last modification by $Author: cxl $)
 
   (C) 1997, Bremen Institute for Safe Systems, Universitaet Bremen
 
  ************************************************************************** *)


signature THEORY =     
    sig
	type theory

	(* starting a proof *)
	val goal : theory* string-> thm list
	(* read a theory *)
	val use  : string-> theory  (* note difference to Isabelle's use *)

	(* theory file name suffix (without leading dot) *)
	val theoryfile_suffix : string

	val theory_of_thm : thm-> theory
	val name_of       : theory-> string
	val eq            : theory* theory-> bool
	val parents_of    : theory-> theory list

	(* the proof state, and saving/restoring it *)
	type proofstate 
	val save          : unit-> proofstate
	val restore       : proofstate-> unit
    end

signature TRAFOTHEORY =
    sig
	(* --- 1. Theories in general (see above), and topthy ------------ *)
       
	structure Thy : THEORY

	val topthy    : Thy.theory

	(* --- 2. Transformational development -------------------------- *)

	(* Transformation inside a context.
	 *
	 * This requires the transformation relation to be monotone wrt. 
	 * the given context. This is stated by thm(s) mono. 
	 * context_var is the variable representing the context, as a 
	 * cterm. prove_mono is a tactic which tries to prove the proof
	 * obligation caused by the mono requirements. This can fail, raising
	 * either ERROR or not decreasing the number of subgoals; in both
	 * cases, apply_trafo_at (below) will raise exception Error.
	 * prove_ext is a tactic which deals with extensionality, used for
	 * transformation under contexts with bound variables. prove_ext should
	 * not produce new subgoals.
	 *)
	val mono        : thm
	val context_var : cterm
	val prove_mono  : int-> tactic
	val prove_ext   : int-> tactic

	(* Transitivity and reflexivity of transformation relation *)
	val trans     : thm list
	val refl      : thm list


	(* --- 3. The basic datatype Transformations -------------------- *)

	(* an instantiation for a transformation is a list of strings
	 * denoting terms *)
	type inst = string list

	type trafo	 = 
	    {(* the list of the names of the parameters. The order is
	      * relevant: the first is the name of the first parameter etc.
	      *)
	     params       : string list, 
	     
	     (* If the transformation wants to instantiate its parameters
	      * other than by the usual
	      * th RS (read_instantiate_sg [("x", ...)] spec) ... RS mp
	      *
	      * This function gets passed a list of parameters, and has
	      * to return a theorem which can be resolved with the current
	      * transformational development. 
	      * This function can be used eg. to check the parameter values.  
	      *)
	     instantiate : (Sign.sg-> inst -> thm) TkTypes.option,
	     	     
	     (* The tactical sugar which is called after the transformation
	      * has been applied. This can discharge proof obligations etc.
	      * The argument is the subgoal corresponding to the current
	      * development.
	      *)
	     sugar       : (int-> tactic) TkTypes.option,
	     
	     (* the logical core theorem. Only actually used if 
	      * instantiate is empty (though of course instantiate may
	      * want to use the core theorem itself).
	      *)
	     core_thm    : thm,

	     (* The name of the trafo, by which it is referred to in 
	      * documentation, on the gen_gui etc. Need not be an ML
	      * identifier. *)
	     name        : string,
	     (* The id of the trafo. This has to be a unique (wrt to all
	      * other transformations) identifier by which the trafo is
	      * externally represented *)
	     id          : string,
	     (* some meaningless waffle about this transformation *)
	     info        : string,
	     (* .. and some explanation about the parameters *)
	     param_info  : string list
	     }

        (* this is the list of all basic trafos known initially *) 
	val all_trafos : unit -> trafo list 

    end

signature TRAFOS =
    sig
	
	type trafo
	type inst   
	
	type path              (* a path into a term *)
	type progdev           (* an ongoing program development *)

        type signatur          (* a collection of operations aka. an Isabelle
				* theory *)
	type spec     = string (* a "specification" aka. a goal text *)

	val start_dev         : signatur* spec-> progdev

	(* applying trafos *)
	val apply             : progdev* trafo* inst-> progdev
	val apply_at          : progdev* trafo*inst* path-> progdev
	    
	(* proof obligations *)
	val can_discharge_thm : progdev* thm* int-> bool
	val discharge_thm     : progdev* thm* int-> progdev
	(* val discharge_simpset : progdev* simpset* int-> progdev *)
	(* get one proof obligation, the number of po's or the whole lot *)
	val get_proof_obl     : progdev* int-> term
	val num_proof_obl     : progdev-> int
	val all_proof_obl     : progdev-> term list


	(* All known trafos. 
	 * Eventually, we want this to be a list of preconceived trafos, plus
	 * a function which takes theorems and (possibly makes them into a 
	 * trafo), eg. fold/unfold like rules. 
	 *)
	val all_trafos       : unit-> trafo list

	(* --- "Meta-development" ----------------------------------------- *)

	(* Create a transformation from given transformational development.
	 * Second argument is the name o/t trafo 
	 *)
	val trafo_from_dev   : progdev* string-> trafo
	    
	(* --- Low-level interface for the GUI. --------------------------- *)

	(* prt_prfstate prints the proofstate in tagged form suitable for
	 * the focus package, with the annotations as paths.
	 *)
	val prt_prfstate     : progdev-> string

	(* Extract current specification from an ongoing program development *)
	val current_spec    : progdev-> term

	(* Extract the initial specification i.e. start o/t development *)
	val initial_spec    : progdev-> term

	(* signature, and theory of an ongoing development *)
	val get_sign         : progdev-> Sign.sg
	val get_theory       : progdev-> signatur
	val top_thm          : progdev-> thm

        (* Something's gone wrong. 
	 * The string contains an informative error message *)
	exception Error of string

    end




functor Trafos(trafos : TRAFOTHEORY) :
    sig include TRAFOS 
	sharing type trafos.trafo      = trafo
	    and type trafos.inst       = inst
	    and type trafos.Thy.theory = signatur
    end =

struct
    
	    
    type inst    = trafos.inst
    type trafo	 = trafos.trafo
    type ext_rep = string
	
    type path     = (* TermSel.path *) string

    type progdev  = trafos.Thy.proofstate
    type signatur = trafos.Thy.theory 
    type spec     = string

    fun debugmsg msg =  Debug.print 22 ("TAS: "^msg)


    exception Error of string 

    (* Extract top theorem, theory and signature from ongoing development *)

    fun top_thm pd = 
	(trafos.Thy.restore pd; topthm())

    fun get_sign pd =
	#sign(rep_thm (top_thm pd))

    fun get_theory pd= trafos.Thy.theory_of_thm (top_thm pd)
	


    (* --- The Focus ------------------------------------------------------ *)
	
    (* The last goal of the current proof state corresponds to
     * the current state of the development; the other goals are the proof 
     * obligations. In general, the current specification has the 
     * following form:
     * 
     *     SP_1 ~~> ?X 
     *
     * which as an Isabelle term looks like this
     *
     * Const (Trueprop, _) $ 
     *   (Const("~~>",  _) $ 
     *        (SP_1 $ Var ("X", _)))
     *
     * The function get_current_goal extracts SP_1 from the given term. The
     * ?X actually isn't important, it can be any term. 
     *)	
    	
    fun get_current_spec (_ $ ((_ $ sp) $ _)) = sp
      | get_current_spec _ = raise (Error "Malformed current specification")
	        
    fun get_context path = 
	let val sg      = #sign(rep_thm(topthm()))
	    val curgoal = getgoal(nprems_of(topthm()))
      	in  TermSel.cut_context sg (get_current_spec (curgoal))
	                           (TermSel.decode path)
	end

    (* splits a list into ist last element and the remaining front bit *)
    fun split []     = raise Subscript
      | split [x]    = ([], x)
      | split (h::t) = let val (h1, t1)= split t in (h::h1, t1) end      

    fun current_spec st = 
	get_current_spec (#2 (split (prems_of(top_thm st))))
	handle Subscript => raise (Error "Malformed specification.")

    fun current_proof_obl st =
	#1 (split (prems_of(top_thm st)))	   
	handle Subscript => raise (Error "Malformed specification.")
    
    fun initial_spec st =
	get_current_spec (concl_of(top_thm st))

    fun prt_prfstate st =
	let val tt = top_thm st
	    val th = theory_of_thm tt
	    val cg = getgoal(nprems_of tt)
	in  TermSel.string_of th (get_current_spec cg)
	end handle Error e=> e (* :-) *)


    (* --- Applying transformations --------------------------------------- *)


    (* the standard way to instantiate parameters *)
    fun inst_param(sg, trf:trafo, i) = 
	case #instantiate trf of
	    SOME inst_fun => inst_fun sg i
	  | NONE=> 
		let (* Standard Lib.List.foldl, Isabelle's foldr would use
                     * fun intro_param(th, str) = ... *)
                    fun intro_param(str, th) =
		        th RS (read_instantiate_sg sg [("x",str)] spec) 
		    val ict = Lib.List.foldl intro_param (#core_thm trf) i
			      (* handle THM _ => th ?? *)
		in  (* if the following forward resolution fails (exn THM),
                     * then the core theorem is not an implication, so we just
		     * leave it (shouldn't happen actually).
		     *)
		    (ict RS mp handle THM _ => ict)
		end	    

    fun applyTrafo (prf, trf, i, p) =
	let val _  = trafos.Thy.restore prf
	    val _  = debugmsg("Applying trafo "^(#name trf)^" ["^
			           (BasicUtil.concatWith ", " i)^"]")
	    val n  = nprems_of(topthm()) 
	    val sg = #sign(rep_thm(topthm()))
	    val t  = inst_param(sg, trf, i)
	    val _  = debugmsg("Instantiated core thm: "^(string_of_thm t))
	(* apply the transformation *)
	    val _  = case p of 
		     SOME pth  => 
			let val ct = get_context pth
			    val ic = Fix.cterm_instantiate 
				         [(trafos.context_var, ct)]
					 trafos.mono
			    val _  = by (resolve_tac trafos.trans n)
			    val _  = by (rtac ic n)
			    val nn = nprems_of (topthm())
			    val _  = by (trafos.prove_ext n)
			    val nx = nprems_of (topthm())
			    val _  = if nx > nn then 
					  (* extensionality tactic failed *)
					  raise Error 
					        "Cannot apply transformation \
                                                \in context\nwith bound \
						\variables.\n"					 
				     else if nn-1 > n then
					 (* we have additional monotonicity
					  * subgoals *)
					 by (trafos.prove_mono (nn-1))
				     else () (* handle ERROR ... ? *)
			in
			    by (rtac t n)
			end
		      | NONE => by ((resolve_tac trafos.trans n)
				    THEN (rtac t n))
	(* apply any tactical sugar *)
	    val _ = case  #sugar trf of
		     SOME tac_sug => by (tac_sug n)			     
		   | NONE => ()
	in  trafos.Thy.save()
	end

    fun apply (pd, trf, i) = 
	applyTrafo (pd, trf, i, NONE)

    fun apply_at (pd, trf, i, p) = 
	applyTrafo (pd, trf, i, SOME p)


    (* --- Proof Obligations --------------------------------------------- *)
	
    (* This checks whether theorem thm discharges the proof obligation 
     * (i.e. makes it go away). 
     * The check isn't terribly sophisticated, it just tests if the number
     * of subgoals decreases.
     *)
    fun can_discharge_thm(prf, thm, subgoal)= 
	let val _    = trafos.Thy.restore prf
	    val befor= nprems_of (topthm()) 
	    val _    = by (rtac thm subgoal)
	    val after= nprems_of (topthm())
	in  after < befor
	end handle ERROR => false

    fun discharge_thm(prf, thm, subgoal)= 
	(trafos.Thy.restore prf;
	 by (rtac thm subgoal); (* handle ERROR .. *)	 
	 trafos.Thy.save())

    fun discharge_simpset(prf, simps, subgoal)=
	(trafos.Thy.restore prf;
	 by (asm_simp_tac subgoal simps); (* handle ERROR .. *)
	 trafos.Thy.save())

    fun all_proof_obl st =
	#1 (split (prems_of(top_thm st)))
	handle Subscript => raise (Error "Malformed specification.")

    fun get_proof_obl(st, num) =
	Lib.List.nth(prems_of(top_thm st), num-1)
	handle Subscript => raise (Error "Malformed specification.")

    fun num_proof_obl st =
	Lib.List.length(prems_of(top_thm st))-1
	handle Subscript => raise (Error "Malformed specification.")


    val all_trafos= trafos.all_trafos
	
    (* --- Start a new proof/transformational development ---------------- *)

    fun start_dev (sign, spec) =
	(trafos.Thy.goal(sign, spec)
	      (* handle ERROR  ... *);
	 trafos.Thy.save())

    (* --- Meta-development: Close a development, make it a trafo --------- *)

    fun trafo_from_dev(pd, nm) = 
	let val _       = trafos.Thy.restore pd
	    val n       = nprems_of(topthm())
	    val _       = by (resolve_tac trafos.refl n)
	    val core_th = uresult()
	in  { params = [],
	      core_thm= core_th, 
	      instantiate= NONE,
	      sugar= NONE,
	      name= nm,
	      id= "generated",
	      info= "This transformation was generated from program development "^nm^"\n",
	      param_info= []
	    }
	end
	 
	
	

end












