(* *********************************************************************** *)
(*									   *)
(* Project: sml/Tk: an Tk Toolkit for sml	 			   *)
(* Author: Burkhart Wolff, University of Bremen	 			   *)
(* Date: 25.7.95				 			   *)
(* Purpose of this file: BASIC PROCESS CONTROL AND COMMUNICATION ROUTINES  *)
(*									   *)
(* *********************************************************************** *)

structure Com : COM
= 
struct

local open BasicTypes in
local open BasicUtil in

val smltk_appid : AppId = "SMLTK";

fun selAppId         ap = let val (id,_,_,_,_,_)    = ap in id   end;
fun selAppProt       ap = let val (_,_,_,prot,_,_)  = ap in prot end;
fun selAppIn         ap = let val (_,_,inst,_,_,_)  = ap in inst  end;
fun selAppOut        ap = let val (_,outs,_,_,_,_)  = ap in outs end;
fun selAppCallBack   ap = let val (_,_,_,_,cb,_)    = ap in cb   end;
fun selAppQuitAction ap = let val (_,_,_,_,_,qa)    = ap in qa   end;

fun getApp appId =
    let
	val (apps,_,_,_,_,_) = !COM_state
	val app = BasicUtil.get ((fn y => appId=y) o selAppId) apps
	    handle GET => raise (TCL_ERROR ("Error in function com.getApp (application id "^appId^")"))
    in
	app
    end;

fun getAppProt       appId = selAppProt(getApp appId);
fun getAppIn         appId = selAppIn(getApp appId);
fun getAppOut        appId = selAppOut(getApp appId);
fun getAppCallBack   appId = selAppCallBack(getApp appId);
fun getAppQuitAction appId = selAppQuitAction(getApp appId);

fun addAppI (app as (appId,tclo,tcli,prot,cb,qa)) = 
    let 
	val (apps,logf,wishp,tclini,srcp,fntp) = !COM_state
    in
	COM_state:=(apps@[app],logf,wishp,tclini,srcp,fntp)
    end;

fun addApp (appId,(prog,parms),prot,cb,qa) =
    let
	val (i1,o1) = SysDep.execute(prog,parms)
	val o2      = open_out(prot)
    in
	addAppI(appId,i1,o1,o2,cb,qa)
    end;


fun removeApp appId =
    let
	val (apps,logf,wishp,tclini,srcp,fntp) = !COM_state
	val napps = BasicUtil.delete ((fn y => appId=y) o selAppId) apps
    in
	((getAppQuitAction(appId))();
	 COM_state:=(napps,logf,wishp,tclini,srcp,fntp))
    end;



val getProt = fn () => getAppProt smltk_appid;
val getIn   = fn () => getAppIn   smltk_appid;
val getOut  = fn () => getAppOut  smltk_appid;


(* very ineficcient. should be realized by a "hook" in order to make *)
(* the protocolling optional. if PROT="", no protocolling should     *)
(* take place *)

(* getLine() strings can only be used for texts that are certain not *)
(* contain \n. Otherwise, getlineM() (M for multiple) has to be used.*)
(* On the other side, an appropriate writeM is provided. *)

fun getLineApp appId = 
    let
	val outs = getAppOut appId
	val prot = getAppProt appId
	val t = input_line(outs);
	(* strip off concluding "\n" *) 
	val t = substring(t,0,(size t) -1)
    in 	
	output(prot, "<== " ^ t ^"\n");
	flush_out(prot);
	t
    end

and getLine () = getLineApp smltk_appid;

fun getLineMApp appId = 
    let
	val outs = getAppOut appId
	val prot = getAppProt appId
	fun getls () = 
	    let
		val t = input_line(outs)
	    in    
		output(prot, "<== " ^ t ^"\n");
		flush_out(prot);
		if t = "EOM\n" then "" else t ^ getls () 
	    end
    in
	getls ()
    end

and getLineM () = getLineMApp smltk_appid;

fun putLineApp appId ps = 
    let
	val inst = getAppIn appId
	val prot = getAppProt appId
    in
	output(prot, "==> " ^ ps ^ "\n"); 			
	flush_out(prot);
	output(inst, ps ^ "\n");
	flush_out(inst)
    end

and putLine ps = putLineApp smltk_appid ps; 



fun putTclCmd cmd =
    let
	val emsg = fn s => (concatWith " " s)

	fun getAnswer aws =
	    let
		val a    = getLine()
		val ss   = words a
		val kind = hd ss
		val _ = Debug.print 1 ("Eventloop.getTclValue: got \""^a^"\"");
	    in
		if (kind = "CMDOK" orelse kind = "ERROR" ) then
		    (a,aws)
		else
		    getAnswer(aws@[a])
	    end

	val _           = putLine ("WriteCmd \"CMDOK\" {"^ cmd ^ "}")
	val (a,binds)   = getAnswer []

	val gaws = getTclAnswersGUI()
	val _    = updTclAnswersGUI(gaws@binds)

	val _    = if not (length binds = 0) then
	               Debug.warning "Missed Binding"
		   else
		       ()
    in
	case (hd (words a)) of
	    "CMDOK" => ()
	  | "ERROR" => raise BasicTypes.TCL_ERROR 
		             ("Com.putCmd: got Tcl Error: \""^ a ^"\"")
	  | s       => raise BasicTypes.TCL_ERROR
		             ("Com.putCmd: got unexpected answer: \""^ s ^"\"")
    end


fun readTclVal req =
    let
	val concatSp = BasicUtil.concatWith " "

	fun getAnswer aws =
	    let
		val a    = getLine()
		val ss   = words a
		val kind = hd ss
		val _ = Debug.print 1 ("Eventloop.getTclValue: got \""^a^"\"");
	    in
		if (kind = "VValue" ) then
		    (concatSp(tl(ss)), aws)
		else
		    getAnswer(aws@[a])
	    end

	val _         = putLine ("WriteSec \"VValue\" {"^ req ^ "}")
	val (a,binds) = getAnswer []

	val gaws = getTclAnswersGUI()
	val _    = updTclAnswersGUI(gaws@binds)
    in
	a
    end





(* Tcl - Primitives  *)
(* communication - routine that sends action-representations
   back to sml for interpretation *)

val commToTcl    = "Write"; (* "communicate"; *)
val writeToTcl   = "Write";
val writeMToTcl  = "WriteM";


val prelude_tcl = 
"proc Write {msg} {                     \n \
\  puts  stdout $msg                    \n \
\  flush stdout                         \n \
\}                                      \n \
\proc WriteSec {tag msg} {              \n \
\  set status [catch {eval $msg} res]   \n \
\  if {$status == 0} {                  \n \
\    puts stdout \"$tag $res\"          \n \
\  } else {                             \n \
\    puts stdout \"ERROR $res\"         \n \
\  }                                    \n \
\  flush stdout                         \n \
\}                                      \n \
\proc WriteCmd {tag msg} {              \n \
\  set status [catch {eval $msg} res]   \n \
\  if {$status == 0} {                  \n \
\    puts stdout \"$tag\"               \n \
\  } else {                             \n \
\    puts stdout \"ERROR $res\"         \n \
\  }                                    \n \
\  flush stdout                         \n \
\}                                      \n \
\proc WriteM {msg} {                    \n \
\  puts  stdout $msg                    \n \
\  flush stdout                         \n \
\  puts  \"EOM\"                        \n \
\  flush stdout                         \n \
\}                                      \n ";


(* Tcl - PROCESS CONTROL  *)

fun initTcl cb = 
    let
	val (i1, o1) = SysDep.execute (getWishPath (),[])
	val o2       = open_out (getLogfilename ())
(*	val qa       = fn () => (putLine "destroy ."; putLine "exit 0")
 *)
	val qa       = fn () => (putLine "destroy .")
    in    
	addAppI (smltk_appid,i1,o1,o2,cb,qa);
	putLine (getTclInit() ^ prelude_tcl);
	(* dummy for not handling exceptions *)
	true 
    end;

fun runTcl ()  = (); (* (putLine "EOM"); *)

(* ### 
   hier muss noch sichergestellt werden, dasz die andren Applikation
   auch gekillt werden
*)

fun exitTcl()  = 
    case !COM_state of
	([],_,_,_,_,_) => ()
      | (_,_,_,_,_,_)  => let
			  fun selApps (apps,_,_,_,_,_) = apps
		      in
			  (app (removeApp o selAppId) (selApps(!COM_state));
			   GUI_state:=([],[],[]);
			   COM_state:= ([],getLogfilename(),getWishPath(),getTclInit(),getSrcPath(),getFntPath()))
		      end;

fun resetTcl()  =  (GUI_state:=([],[],[]);
		     COM_state:= ([],getLogfilename (),
				  getWishPath(),getTclInit(),getSrcPath(),getFntPath()));

end; end;

end;

